
 #
 # Copyright (c) 2015, University of Maryland All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without modification,
 # are permitted provided that the following conditions are met:
 #
 # 1. Redistributions of source code must retain the above copyright notice, this
 # list of conditions and the following disclaimer.
 #
 # 2. Redistributions in binary form must reproduce the above copyright notice,
 # this list of conditions and the following disclaimer in the documentation and/or
 # other materials provided with the distribution.
 #
 # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 # ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 # ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #

from shapely.geometry import Point
from shapely.geometry import MultiPoint
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon
import logging
import socket
import sys
import os

sys.path.append(os.path.abspath('../target_region_computation'))

import regions as r

MILES_PER_METER = 0.000621371        #miles/meter

# Get the physical distance from a point to a multipolygon
def getPhysicalRadius(lon, lat, mpoly):
	return r.dist_point_to_mpoly_nb(mpoly, lon, lat)*MILES_PER_METER #miles

# Convert physical distance to latency
def getLatencyRadius(distance):
	#SLOPE = 0.0269 					#ms/mile -- the Htrae slope
	#SLOPE = 0.01345                    #ms/mile -- 1/2 of the Htrae slope
	SLOPE = 0.0161                    	#ms/mile -- 3/c [Equation 4]

	#INTERCEPT = 63.32					#ms -- the Htrae intercept
	INTERCEPT = 0.0

	return distance*SLOPE + INTERCEPT   #ms

def get_next_hop(query_id, cur_name, src_lon, src_lat, dst_lon, dst_lat, avoidZone, relayZone, cur_lon, cur_lat, n_list, nn_list, prev_node=None, path=[], STRATEGY=1):
	# Compute the great circle distance between the node itself and the closest point in F [Section 4.4.1].
	avoid_physical_radius = getPhysicalRadius(cur_lon, cur_lat, avoidZone)

	# Compute the estimate of the minimum possible RTT between the node itself and any possible host in forbidden region F [Section 4.4.1]
	latency_radius = getLatencyRadius(avoid_physical_radius)

	relay_physical_radius = getPhysicalRadius(cur_lon, cur_lat, relayZone)

	safe_nodes = []
	# STRATEGY 0 is the strategy described in the paper [Section 4.4].
	if STRATEGY == 0:
		logging.debug('QUERYID %d;STRATEGY %d;avoid_physical %f;relay_physical %f;latency %f;progress - %f;lon - %f;lat - %f', query_id, STRATEGY, avoid_physical_radius, relay_physical_radius, latency_radius, relay_physical_radius - avoid_physical_radius, cur_lon, cur_lat, extra={'hostname':cur_name})
		
		for neighbor in n_list:
			(neighbor_id, neighbor_lon, neighbor_lat, neighbor_rtt) = neighbor

			if prev_node != None and neighbor_id == prev_node:
				continue

			# Check that the neighbor satisfies the safety condition [Equation 3].
			if float(neighbor_rtt) < latency_radius: 
				# Compute the great cirle distance between the neighbor and the closest point in F and T, respectively [Section 4.4.2]
				avoid_physical_radius2 = getPhysicalRadius(neighbor_lon, neighbor_lat, avoidZone)
				relay_physical_radius2 = getPhysicalRadius(neighbor_lon, neighbor_lat, relayZone)

				# Compute the progress the neighbor makes [Section 4.4.2].
				l_min_distance = relay_physical_radius2 - avoid_physical_radius2

				latency_radius2 = getLatencyRadius(avoid_physical_radius2)

				# Make use of data about two-hop neighbors [Section 4.5].
				for neighbor2 in nn_list.get(neighbor_id,[]):
					(neighbor_id2, neighbor_lon2, neighbor_lat2, neighbor_rtt2) = neighbor2

					# Check the safety condition.
					if float(neighbor_rtt2) < latency_radius2:
						# Compute the progress.
						distance=getPhysicalRadius(neighbor_lon2, neighbor_lat2, relayZone)-getPhysicalRadius(neighbor_lon2, neighbor_lat2, avoidZone)

						if distance < l_min_distance:
							l_min_distance = distance

				# Return IDs and the progress of all the safe nodes.
				safe_nodes.append((neighbor_id, l_min_distance))
	# STRATEGY 1 is a strategy that does not use "query forking" [Section 4.5].
	elif STRATEGY == 1:
		# Compute the progress of the current node
		relay_physical_radius = getPhysicalRadius(cur_lon, cur_lat, relayZone)
		min_distance = relay_physical_radius - avoid_physical_radius
		min_node_id = None

		logging.debug('QUERYID %d;STRATEGY %d;avoid_physical %f;relay_physical %f;latency %f;progress - %f;lon - %f;lat - %f', query_id, STRATEGY, avoid_physical_radius, relay_physical_radius, latency_radius, relay_physical_radius - avoid_physical_radius, cur_lon, cur_lat, extra={'hostname':cur_name})

		for neighbor in n_list:
			(neighbor_id, neighbor_lon, neighbor_lat, neighbor_rtt) = neighbor

			if prev_node != None and neighbor_id == prev_node:
				continue

			# Check the safety condition.
			if float(neighbor_rtt) < latency_radius:
				avoid_physical_radius2 = getPhysicalRadius(neighbor_lon, neighbor_lat, avoidZone)
				relay_physical_radius2 = getPhysicalRadius(neighbor_lon, neighbor_lat, relayZone)
				l_min_distance = relay_physical_radius2 - avoid_physical_radius2

				latency_radius2 = getLatencyRadius(avoid_physical_radius2)

				for neighbor2 in nn_list.get(neighbor_id, []):
					(neighbor_id2, neighbor_lon2, neighbor_lat2, neighbor_rtt2) = neighbor2

					if float(neighbor_rtt2) < latency_radius2:
						distance=getPhysicalRadius(neighbor_lon2, neighbor_lat2, relayZone)-getPhysicalRadius(neighbor_lon2, neighbor_lat2, avoidZone)

						if distance < l_min_distance:
							l_min_distance = distance

				# Find a neighbor that makes the most progress.
				if l_min_distance < min_distance:
					min_node_id = neighbor_id
					min_distance = l_min_distance

		if min_node_id == None:
			logging.info('QUERYID %d;STRATEGY %d;Failed to find the next hop;', query_id, STRATEGY, extra={'hostname':cur_name})

		# Return the neighbor that makes the most progress.
		safe_nodes.append(min_node_id)
	return safe_nodes

def check_nlist_within(query_id, cur_name, cur_lon, cur_lat, n_list, nn_list, avoidZone, relayZone):
	avoid_physical_radius = getPhysicalRadius(cur_lon, cur_lat, avoidZone)          # miles
	relay_physical_radius = getPhysicalRadius(cur_lon, cur_lat, relayZone)          # miles

	logging.info('QUERYID %d;Checking if any neighbor is inside target region', query_id, extra={'hostname':cur_name})

	#distance = relay_physical_radius - avoid_physical_radius
	#if distance > 0:
	#	return None

	latency_radius = getLatencyRadius(avoid_physical_radius)                # ms

	for neighbor in n_list:
		(neighbor_id, neighbor_lon, neighbor_lat, neighbor_rtt) = neighbor

		if float(neighbor_rtt) < latency_radius: 
				
			if Point(neighbor_lon, neighbor_lat).intersects(relayZone):
				return neighbor_id


			avoid_physical_radius2 = getPhysicalRadius(neighbor_lon, neighbor_lat, avoidZone)          # miles
			relay_physical_radius2 = getPhysicalRadius(neighbor_lon, neighbor_lat, relayZone)          # miles

			latency_radius2 = getLatencyRadius(avoid_physical_radius2)                # ms

			if not nn_list.has_key(neighbor_id):
				logging.info('Do not have neighbor list of %s;', neighbor_id, extra={'hostname':cur_name})
				continue

			#continue

			for neighbor2 in nn_list[neighbor_id]:
				(neighbor_id2, neighbor_lon2, neighbor_lat2, neighbor_rtt2) = neighbor2

				if float(neighbor_rtt2) < latency_radius2:
						
					if Point(neighbor_lon2, neighbor_lat2).intersects(relayZone):
						return neighbor_id2


	return None

def is_in_path(t_ip, path):
	t_node = socket.gethostbyaddr(t_ip)
	for node in path:
		if t_node == node[0]:
			return True

	return False


