#!/usr/bin/env python

 #
 # Copyright (c) 2015, University of Maryland All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without modification,
 # are permitted provided that the following conditions are met:
 #
 # 1. Redistributions of source code must retain the above copyright notice, this
 # list of conditions and the following disclaimer.
 #
 # 2. Redistributions in binary form must reproduce the above copyright notice,
 # this list of conditions and the following disclaimer in the documentation and/or
 # other materials provided with the distribution.
 #
 # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 # ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 # ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #

import pickle
import marshal
import struct
from shapely.wkb import dumps as b_dumps, loads as b_loads
from shapely.wkt import dumps as t_dumps, loads as t_loads
from zlib import compress, decompress
import socket
import hmac
from hashlib import sha1


# types for struct

#	Format	C_Type				Python_type
#	  x	   pad byte				no value
#	  c	   char 				string of length 1
#	  b	   signed char 			integer
#	  B	   unsigned char 		integer
#	  ?	   _Bool 				bool
#	  h	   short 				integer
#	  H	   unsigned short		integer
#	  i	   int 					integer
#	  I	   unsigned int 		integer
#	  l	   long 				integer
#	  L	   unsigned long 		integer
#	  q	   long long 			integer
#	  Q	   unsigned long long 	integer
#	  f	   float 				float
#	  d	   double 				float
#	  s	   char[] 				string
#	  p	   char[] 				string
#	  P	   void * 				integer


# constants for msg types

QUERY_MSG=0
JOIN_MSG=1
GOSSIP_MSG=2
REQUEST_NLIST_MSG=3
REQUEST_QUERY_MSG=4
QUERY_RESPONSE_MSG=5
DUP_CHECK_REQUEST_MSG=6
DUP_CHECK_RESPONSE_MSG=7
QUERY_OVERHEAD_REQUEST_MSG=8
QUERY_OVERHEAD_RESPONSE_MSG=9
QUERY_CHECK_REQUEST_MSG=10
QUERY_CHECK_RESPONSE_MSG=11
PING_MSG=12
PONG_MSG=13



class Msg:
	def __init__(self, arg_list):
		self.arg_list = arg_list

	def marshal(self):
		value_list=[]
		fmt_str=''

		#buf = None
		buf = ''

		#print self.arg_list

		# assume that vars always come after the supported formats
		for v1,v2 in self.arg_list:
			if v2=='var':
				#if buf == None: buf=struct.pack(fmt_str, *value_list)
				var_len = len(v1)
				buf+=struct.pack('i', var_len)
				buf+=v1
				continue
			buf+=struct.pack(v2, v1)
			#value_list.append(v1)
			#fmt_str+=v2

		#if buf == None: buf=struct.pack(fmt_str, *value_list)

		#print fmt_str
		#print value_list
	
		#print struct.unpack(fmt_str, buf[:struct.calcsize(fmt_str)])

		#print len(buf)

		return struct.pack('i', len(buf))+buf

	def unmarshal(self, buf):
		fmt_str=''

		buf = buf[struct.calcsize('i'):]
		#buf = buf[struct.calcsize('i')+20:]
		ret = ()
		offset = 0

		for v1, v2 in self.arg_list:
			if v2=='var': continue
			fmt_str+=v2
			ret+=struct.unpack(v2, buf[offset:offset+struct.calcsize(v2)])
			offset+=struct.calcsize(v2)

		#buf = buf[4:]

		#offset=struct.calcsize(fmt_str)
		buf_len=len(buf)
		#ret=struct.unpack(fmt_str, buf[:offset])


		while offset < buf_len:
			(var_len,)=struct.unpack('i', buf[offset:offset+struct.calcsize('i')])
			#print offset, fmt_str, var_len
			ret+=(buf[offset+struct.calcsize('i'):offset+struct.calcsize('i')+var_len],)
			offset+=(struct.calcsize('i')+var_len)

		return ret[1:]

class RequestNlistMsg (Msg):
	def __init__(self, lon=None, lat=None):
		arg_list = []
		arg_list.append((REQUEST_NLIST_MSG,'i'))
		arg_list.append((lon,'d'))
		arg_list.append((lat,'d'))


		Msg.__init__(self, arg_list)


class PingMsg (Msg):
	def __init__(self, nonce=None):
		arg_list = []
		arg_list.append((PING_MSG,'i'))
		arg_list.append((nonce,'I'))
		Msg.__init__(self, arg_list)


class PongMsg (Msg):
	def __init__(self, nonce=None):
		arg_list = []
		arg_list.append((PONG_MSG,'i'))
		arg_list.append((nonce,'I'))
		Msg.__init__(self, arg_list)


class GossipMsg (Msg):
	def __init__(self, nlist=None):
		arg_list = []
		arg_list.append((GOSSIP_MSG,'i'))
		arg_list.append((nlist,'var'))
		Msg.__init__(self, arg_list)


class RequestQueryMsg (Msg):
	def __init__(self, query_id=None, src_lon=None, src_lat=None, dst_lon=None, dst_lat=None, ttl=None, strategy=None, fr=None, tr=None, src_host_name=None, dst_host_name=None):
		arg_list = []
		arg_list.append((REQUEST_QUERY_MSG,'i'))
		arg_list.append((query_id,'i'))
		arg_list.append((src_lon,'d'))
		arg_list.append((src_lat,'d'))
		arg_list.append((dst_lon,'d'))
		arg_list.append((dst_lat,'d'))
		arg_list.append((ttl,'i'))
		arg_list.append((strategy,'i'))
		if fr==None:
			arg_list.append((None,'var'))
		else:
			arg_list.append((b_dumps(fr),'var'))
		if tr==None:
			arg_list.append((None,'var'))
		else:
			arg_list.append((b_dumps(tr),'var'))
		arg_list.append((src_host_name,'var'))
		arg_list.append((dst_host_name,'var'))

		Msg.__init__(self, arg_list)

	def unmarshal(self, buf):
		(query_id, src_lon, src_lat, dst_lon, dst_lat, ttl, strategy, b_fr, b_tr, src_host_name, dst_host_name) = Msg.unmarshal(self, buf)
		return (query_id, src_lon, src_lat, dst_lon, dst_lat, ttl, strategy, b_loads(b_fr), b_loads(b_tr), src_host_name, dst_host_name)


	'''
	def __init__(self, grid_cell_size=None, ineq_factor=None, tolerance=None, dst=None, *args):
		arg_list = []
		arg_list.append((REQUEST_QUERY_MSG,'i'))
		arg_list.append((grid_cell_size,'d'))
		arg_list.append((ineq_factor,'d'))
		arg_list.append((tolerance,'d'))
		arg_list.append((dst,'var'))
		all_fc=''
		for fc in args:
			#note that country code is always 3 bytes
			all_fc+=fc
		arg_list.append((all_fc,'var'))

		Msg.__init__(self, arg_list)


	def unmarshal(self, buf):
		#note that country code is always 3 bytes
		(grid_cell_size, ineq_factor, tolerance, dst, all_fc) = Msg.unmarshal(self, buf)

		fc_list=[]
		for i in xrange(0,len(all_fc),3):
			fc_list.append(all_fc[i:i+3])

		return (grid_cell_size, ineq_factor, tolerance, dst, fc_list)
	'''

class JoinMsg (Msg):
	def __init__(self):
		arg_list = []
		arg_list.append((JOIN_MSG,'i'))

		Msg.__init__(self, arg_list)


class DupCheckRequestMsg (Msg):
	def __init__(self, query_id=None):
		arg_list = []
		arg_list.append((DUP_CHECK_REQUEST_MSG,'i'))
		arg_list.append((query_id,'i'))

		Msg.__init__(self, arg_list)

class QueryCheckRequestMsg (Msg):
	def __init__(self, query_id=None):
		arg_list = []
		arg_list.append((QUERY_CHECK_REQUEST_MSG,'i'))
		arg_list.append((query_id,'i'))

		Msg.__init__(self, arg_list)

class DupCheckReplyMsg (Msg):
	def __init__(self, query_id=None, is_dup=0):
		arg_list = []
		arg_list.append((DUP_CHECK_RESPONSE_MSG,'i'))
		arg_list.append((query_id,'i'))
		arg_list.append((is_dup,'i'))

		Msg.__init__(self, arg_list)

class QueryCheckReplyMsg (Msg):
	def __init__(self, query_id=None, is_dup=0):
		arg_list = []
		arg_list.append((QUERY_CHECK_RESPONSE_MSG,'i'))
		arg_list.append((query_id,'i'))
		arg_list.append((is_dup,'i'))

		Msg.__init__(self, arg_list)


class QueryOverheadRequestMsg (Msg):
	def __init__(self):
		arg_list = []
		arg_list.append((QUERY_OVERHEAD_REQUEST_MSG,'i'))

		Msg.__init__(self, arg_list)

class QueryOverheadResponseMsg (Msg):
	def __init__(self, query_overhead=None):
		arg_list = []
		arg_list.append((QUERY_OVERHEAD_RESPONSE_MSG,'i'))
		arg_list.append((query_overhead,'var'))

		Msg.__init__(self, arg_list)


class QueryResponseMsg (Msg):
	def __init__(self, query_id=None, succeeded=0, res_msg=None, path=None):
		arg_list = []
		arg_list.append((QUERY_RESPONSE_MSG,'i'))
		arg_list.append((query_id,'i'))
		arg_list.append((succeeded,'i'))
		arg_list.append((res_msg,'var'))
		arg_list.append((path,'var'))

		Msg.__init__(self, arg_list)


class QueryMsg (Msg):
	def __init__(self, query_id=None, src_lon=None, src_lat=None, dst_lon=None, dst_lat=None, ttl=None, strategy=None, fr=None, tr=None, path=None, src_host_name=None, dst_host_name=None):
		arg_list = []
		arg_list.append((QUERY_MSG,'i'))
		arg_list.append((query_id,'i'))
		arg_list.append((src_lon,'d'))
		arg_list.append((src_lat,'d'))
		arg_list.append((dst_lon,'d'))
		arg_list.append((dst_lat,'d'))
		arg_list.append((ttl,'i'))
		arg_list.append((strategy,'i'))
		if fr==None:
			arg_list.append((None,'var'))
		else:
			arg_list.append((b_dumps(fr),'var'))
		if tr==None:
			arg_list.append((None,'var'))
		else:
			arg_list.append((b_dumps(tr),'var'))
		arg_list.append((path,'var'))
		arg_list.append((src_host_name,'var'))
		arg_list.append((dst_host_name,'var'))

		Msg.__init__(self, arg_list)

	def unmarshal(self, buf):
		(query_id, src_lon, src_lat, dst_lon, dst_lat, ttl, strategy, b_fr, b_tr, path, src_host_name, dst_host_name) = Msg.unmarshal(self, buf)
		return (query_id, src_lon, src_lat, dst_lon, dst_lat, ttl, strategy, b_loads(b_fr), b_loads(b_tr), path, src_host_name, dst_host_name)


class TestMsg (Msg):
	def __init__(self, t=None, i=None, f=None, ss=None, bb=None, nlist=None):
		self.arg_list = []
		self.arg_list.append((10,'b'))
		self.arg_list.append((t,'b'))
		self.arg_list.append((i,'i'))
		self.arg_list.append((f,'f'))
		self.arg_list.append((ss,'var'))
		self.arg_list.append((bb,'var'))
		self.arg_list.append((nlist,'var'))


