#!/usr/bin/env python

 #
 # Copyright (c) 2015, University of Maryland All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without modification,
 # are permitted provided that the following conditions are met:
 #
 # 1. Redistributions of source code must retain the above copyright notice, this
 # list of conditions and the following disclaimer.
 #
 # 2. Redistributions in binary form must reproduce the above copyright notice,
 # this list of conditions and the following disclaimer in the documentation and/or
 # other materials provided with the distribution.
 #
 # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 # ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 # ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #

import socket
import struct
import subprocess
import os
from shapely.geometry import MultiPoint, Point
from random import sample, randint, seed
from math import ceil, atan2, degrees
import numpy as np
#import matplotlib.pyplot as plt
#import statsmodels.api as sm
import forwarding as fwd
from msg import PingMsg, PongMsg
from alibid_cfg import ALIBI_PORT
import util as u
import logging

class Neighbors:
	def __init__(self):
		self.n_list = []
		self.c_list = []
		self.nn_list = {}
		self.n_list_max_len = 32 	# a list of neighbors
		self.c_list_max_len = 128	# a list of known-active peers
		self.gossip_num = 5
		self.ewma_weight = 0.7
		#seed(1)

	# measure and add RTTs to the neighbor list
	def add_latency(self, node_list, random_loss=False):
		up_list = []
		down_list = []
		probes_num = 1
		for node in node_list:
	 		latency = self.measure_latency(node[0], probes_num)
			if random_loss:
				if randint(0,4) < 1:
					latency = None	
			#print node[0], node[1], node[2], latency
			if latency == None:
				down_list.append((node[0],))
			else:
				up_list.append((node[0], node[1], node[2], latency))

		return (up_list, down_list)

	# update latency information of neighbors
	def update_rtts(self, target_list, info):
		for new_measure in info:
			for i in xrange(len(target_list)):
				if new_measure[0] == target_list[i][0]:
					# update using exponentially-weighted moving average
					new_rtt = self.ewma_weight*float(new_measure[3]) + (1-self.ewma_weight)*float(target_list[i][3])
					target_list[i] = (target_list[i][0], target_list[i][1], target_list[i][2], new_rtt)
					break


	# filter out down nodes and remove duplicates
	def process_rtt_result(self, node_lists):
		(up_list, down_list) = node_lists

		# remove downed hosts from the n and c list
		self.n_list[:] = [x for x in self.n_list if not self.is_in_list(x, down_list)]
		self.c_list[:] = [x for x in self.c_list if not self.is_in_list(x, down_list)]

		self.update_rtts(self.n_list, up_list)
		self.update_rtts(self.c_list, up_list)

		# remove duplicates from up nodes
		res = [x for x in up_list if not self.is_in_list(x, self.c_list+self.n_list)]

		return res

	# add nodes to known-active peers
	def add_to_candidate_list(self, node_lists):
		up_list = self.process_rtt_result(node_lists)

		# FIFO
		self.c_list+=up_list
		c_list_len = len(self.c_list)
		if c_list_len > self.c_list_max_len:
			self.c_list = self.c_list[c_list_len-self.c_list_max_len:]

	# update the list of neighbors to maintain diversity
	def update_neighbor_list(self):
		a_list = []

		if len(self.n_list) < self.n_list_max_len:
			a_list = sample(self.c_list, min(self.n_list_max_len - len(self.n_list), len(self.c_list)))
			self.c_list[:]=[x for x in self.c_list if not x in a_list]
			self.n_list+=a_list


		if len(self.n_list) == 0:
			return
		
		r_list = []
		iter_num = randint(3,6)
		for i in xrange(iter_num):
			if len(self.c_list) == 0:
				break

			self.gen_bin_intervals()
			self.fill_bins()

			# choose a neighbor to remove
			node_r = self.choose_one_from_nlist()
			node_a = sample(self.c_list, 1)[0]
			r_list.append(node_r)
			self.c_list.remove(node_a)
			self.n_list.remove(node_r)
			self.n_list.append(node_a)

			a_list.append(node_a)
			self.nn_list.pop(node_a[0], None)

		self.c_list+=r_list

		return a_list

		'''
		#Random
		self.n_list += self.c_list

		if len(self.n_list) > self.n_list_max_len:
			self.n_list[:] = sample(self.n_list, self.n_list_max_len)
		self.c_list[:] = [n for n in self.c_list if n not in self.n_list]
		'''

	# choose one neighbor to remove
	def choose_one_from_nlist(self):
		max_len = None

		# choose the most redundant bin
		for b in self.bins:
			if max_len == None or (b != None and max_len <= len(b)):
				max_len = len(b)
				max_bin = b

		# choose one from max_bin using GPS coords (and delete it)
		target = max_bin[self.angle_redundancy(max_bin)]

		#print target
		return target


	# generate bins based on RTTs
	# strategy: 0 - uniform, 1 - more nearby nodes
	def gen_bin_intervals(self, bin_num=10, strategy=0):
		#uniform
		min_rtt=None
		max_rtt=None

		for (ip, lon, lat, rtt) in self.n_list:
			if min_rtt == None or min_rtt > float(rtt):
				min_rtt = float(rtt)
				
			if max_rtt == None or max_rtt < float(rtt):
				max_rtt = float(rtt)

		interval = ceil((max_rtt-min_rtt)/bin_num)

		self.bin_intervals = []
		split_point = ceil(bin_num/3)
	
		for i in xrange(bin_num+1):
			self.bin_intervals.append(min_rtt+i*interval)
			if strategy == 1 and i < split_point:
				self.bin_intervals.append(min_rtt+(i+0.5)*interval)

		self.bin_intervals[-1]+=0.01

	def fill_bins(self):
		self.bins = [None]*(len(self.bin_intervals)-1)
	
		rtt_list = [rtt for (ip, lon, lat, rtt) in self.n_list]
		bin_indices = np.digitize(rtt_list, self.bin_intervals)

		for i in xrange(len(bin_indices)):
			idx = bin_indices[i]-1
			if self.bins[idx] == None: self.bins[idx] = []
			self.bins[idx].append(self.n_list[i])

	# find a neighbor with the most redundant bearing [Section 4.3 of the paper].
	def angle_redundancy(self, node_info):
		if len(node_info) < 2:
			return randint(0, len(node_info)-1)

		# node_info is a list of node information: ip, coord, latency
		angles = []
		for p in node_info:
			angle=degrees(atan2(float(p[2])-float(self.gps_coords[1]), float(p[1])-float(self.gps_coords[0])))
			if angle < 0: angle+=360
			angles.append((angle, p[0]))

		angles.sort()

		for i in xrange(len(angles)):
			if i == 0: angle = angles[1][0] + 360 - angles[-1][0]
			elif i == (len(angles)-1): angle = angles[0][0] + 360 - angles[-2][0]
			else: angle = angles[i+1][0] - angles[i-1][0]

			angles[i]+=(angle,)

		min_node_ip = min(angles, key = lambda a:a[2])[1]

		for i in range(len(node_info)):
			if node_info[i][0] == min_node_ip:
				return i

		return None


	# serialize neighbors
	def marshal(self, t_list=None):
		if t_list == None: t_list = self.n_list

		ret=''
		for neighbor in t_list:
			ret+=socket.inet_aton(neighbor[0])
			ret+=struct.pack('dd', float(neighbor[1]), float(neighbor[2]))

		return ret

	# serialize neighbors with latency information
	def marshal_wl(self, t_list=None):
		if t_list == None: t_list = self.n_list

		ret=''
		for neighbor in t_list:
			ret+=socket.inet_aton(neighbor[0])
			ret+=struct.pack('ddd', float(neighbor[1]), float(neighbor[2]), float(neighbor[3]))

		return ret

	def unmarshal(self, buf):
		offset = 0
		buf_len = len(buf)
		ret_list = []
		while offset < buf_len:
			ip_str=socket.inet_ntoa(buf[offset:offset+4])
			offset+=4
			fmt_len = struct.calcsize('dd')
			(lon, lat) = struct.unpack('dd', buf[offset:offset+fmt_len])
			offset+=fmt_len
			ret_list.append((ip_str, lon, lat))

		return ret_list

	def unmarshal_wl(self, buf):
		offset = 0
		buf_len = len(buf)
		ret_list = []
		while offset < buf_len:
			ip_str=socket.inet_ntoa(buf[offset:offset+4])
			offset+=4
			fmt_len = struct.calcsize('ddd')
			(lon, lat, rtt) = struct.unpack('ddd', buf[offset:offset+fmt_len])
			offset+=fmt_len
			ret_list.append((ip_str, lon, lat, rtt))

		return ret_list

	# measure RTTs using ping with nonce	
	def measure_latency(self,address,probes_num):
		nonce = int(os.urandom(struct.calcsize('I')).encode('hex'),16)
		ping_msg = PingMsg(nonce)
		(response, latency) = u.send_msg(address, ALIBI_PORT, ping_msg, True, True, True, 1)

		pong_msg = PongMsg()
		(pong_nonce,) = pong_msg.unmarshal(response)

		logging.debug('Checking if nonces are matching (%d == %d)', nonce, pong_nonce, extra={'hostname':socket.gethostbyaddr(socket.gethostname())[0]})

		if nonce == pong_nonce:
			return latency
		else:
			return None

	# store GPS coords (which is used for calculating bearing).
	def set_gps_coords(self, gps_coords):
		self.gps_coords = gps_coords

	def remove_item_from_nlist(self, ip):
		self.nn_list.pop(ip, None)
		self.n_list[:] = [x for x in self.n_list if not self.has_same_ip(x, (ip,))]


	def remove_item_from_list(self, ip, l):
		return [x for x in l if not self.has_same_ip(x, (ip,))]

	def is_in_list(self, item, t_list):
		for t_item in t_list:
			if t_item[0] == item[0]:
				return True

		return False

	def has_same_ip(self, item1, item2):
		if item1[0] == item2[0]:
			return True
		else:
			return False




